import numpy as np
from collections import deque
import os
import os.path as osp
import copy
import torch
import torch.nn.functional as F
import copy

from PromptTrack.tracker.kalman_filter import KalmanFilter
from PromptTrack.tracker import  matching
from PromptTrack.tracker.basetrack import BaseTrack, TrackState

class STrack(BaseTrack):
    shared_kalman = KalmanFilter()
    def __init__(self, tlwh, score, id_in_frame=None, frame_id=None):
        # wait activate
        self._tlwh = np.asarray(tlwh, dtype=np.float64)
        self.kalman_filter = None
        self.mean, self.covariance = None, None
        self.is_activated = False
        self.track_id = None
        self.score = score
        self.tracklet_len = 0
        self.infos ={"frame_id":frame_id, "id_in_frame":id_in_frame, "location":tlwh.tolist(),"detection_score":self.score, "track_id":self.track_id}


    def predict(self):
        mean_state = self.mean.copy()
        if self.state != TrackState.Tracked:
            mean_state[7] = 0
        self.mean, self.covariance = self.kalman_filter.predict(mean_state, self.covariance)

    @staticmethod
    def multi_predict(stracks):
        if len(stracks) > 0:
            multi_mean = np.asarray([st.mean.copy() for st in stracks])
            multi_covariance = np.asarray([st.covariance for st in stracks])
            for i, st in enumerate(stracks):
                if st.state != TrackState.Tracked:
                    multi_mean[i][7] = 0
            multi_mean, multi_covariance = STrack.shared_kalman.multi_predict(multi_mean, multi_covariance)
            for i, (mean, cov) in enumerate(zip(multi_mean, multi_covariance)):
                stracks[i].mean = mean
                stracks[i].covariance = cov

    def activate(self, kalman_filter, frame_id):
        """Start a new tracklet"""
        self.kalman_filter = kalman_filter
        self.track_id = self.next_id()
        self.infos["track_id"]=self.track_id
        self.mean, self.covariance = self.kalman_filter.initiate(self.tlwh_to_xyah(self._tlwh))

        self.tracklet_len = 0
        self.state = TrackState.Tracked
        if frame_id == 1:
            self.is_activated = True
        # self.is_activated = True
        self.frame_id = frame_id
        self.start_frame = frame_id

    def re_activate(self, new_track, frame_id, new_id=False):
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, self.tlwh_to_xyah(new_track.tlwh)
        )
        self.tracklet_len = 0
        self.state = TrackState.Tracked
        self.is_activated = True
        self.frame_id = frame_id
        if new_id:
            self.track_id = self.next_id()
            self.infos["track_id"]=self.track_id
        self.score = new_track.score

    def update(self, new_track, frame_id,id_in_frame=None):
        """
        Update a matched track
        :type new_track: STrack
        :type frame_id: int
        :type update_feature: bool
        :return:
        """
        self.id_in_frame =int(id_in_frame)
        self.frame_id = frame_id
        self.tracklet_len += 1

        new_tlwh = new_track.tlwh
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, self.tlwh_to_xyah(new_tlwh))
        self.state = TrackState.Tracked
        self.is_activated = True
        self.infos ={"track_id":self.track_id,"frame_id":self.frame_id, "id_in_frame":int(id_in_frame), "location":new_tlwh.tolist(),"detection_score":self.score}

        self.score = new_track.score

    @property
    # @jit(nopython=True)
    def tlwh(self):
        """Get current position in bounding box format `(top left x, top left y,
                width, height)`.
        """
        if self.mean is None:
            return self._tlwh.copy()
        ret = self.mean[:4].copy()
        ret[2] *= ret[3]
        ret[:2] -= ret[2:] / 2
        return ret

    @property
    # @jit(nopython=True)
    def tlbr(self):
        """Convert bounding box to format `(min x, min y, max x, max y)`, i.e.,
        `(top left, bottom right)`.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    @staticmethod
    # @jit(nopython=True)
    def tlwh_to_xyah(tlwh):
        """Convert bounding box to format `(center x, center y, aspect ratio,
        height)`, where the aspect ratio is `width / height`.
        """
        ret = np.asarray(tlwh).copy()
        ret[:2] += ret[2:] / 2
        ret[2] /= ret[3]
        return ret

    def to_xyah(self):
        return self.tlwh_to_xyah(self.tlwh)

    @staticmethod
    # @jit(nopython=True)
    def tlbr_to_tlwh(tlbr):
        ret = np.asarray(tlbr).copy()
        ret[2:] -= ret[:2]
        return ret

    @staticmethod
    # @jit(nopython=True)
    def tlwh_to_tlbr(tlwh):
        ret = np.asarray(tlwh).copy()
        ret[2:] += ret[:2]
        return ret

    def __repr__(self):
        return 'OT_{}_({}-{})'.format(self.track_id, self.start_frame, self.end_frame)


class BYTETracker(object):
    def __init__(self, track_thresh=0.45, fixed_parc=False, track_buffer=25, match_thresh=0.8, frame_rate=30,mot20=True,max_time_lost=100,nbr_frames_fixing=10):
        self.tracked_stracks = []  # type: list[STrack]
        self.lost_stracks = []  # type: list[STrack]
        self.removed_stracks = []  # type: list[STrack]
        
        #sophie: 
        self.DBN_previous_track=[]
        self.DBN_current_track=[]

        self.frame_id = 0
        self.track_thresh=track_thresh
        self.track_buffer=track_buffer
        self.match_thresh=match_thresh
        self.frame_rate=frame_rate
        self.mot20=mot20
        self.fixed_parc=fixed_parc
        self.nbr_frames_fixing= nbr_frames_fixing     ###number of frames for initializing animals    
        #self.args = {"track_thresh":track_thresh,"track_buffer":track_buffer,"match_thresh":match_thresh,"frame_rate":frame_rate,"mot20":False}
        #self.det_thresh = args.track_thresh
        self.det_thresh = track_thresh 
        self.buffer_size = int(frame_rate / 30.0 * track_buffer)
        self.max_time_lost = max_time_lost #self.buffer_size ##sophie mod
        self.kalman_filter = KalmanFilter()

    def update(self, output_results, img_info=None, img_size=None):
        self.frame_id += 1
        activated_starcks = []
        refind_stracks = []
        lost_stracks = []
        removed_stracks = []

        if output_results.shape[1] == 5:
            scores = output_results[:, 4]
            bboxes = output_results[:, :4]
        else:
            output_results = output_results.cpu().numpy()
            scores = output_results[:, 4] * output_results[:, 5]  #sophie que veux dire cette nouvelle sortie 
            bboxes = output_results[:, :4]  # x1y1x2y2
        #img_h, img_w = img_info[0], img_info[1]
        scale = 1# min(img_size[0] / float(img_h), img_size[1] / float(img_w))
        bboxes /= scale

        remain_inds = scores > self.track_thresh
        inds_low = scores > 0.0 #Sophie mod 0.1
        inds_high = scores < self.track_thresh
        #print("****",inds_high, inds_low, remain_inds)

        inds_second = np.logical_and(inds_low, inds_high)
        dets_second = bboxes[inds_second]
        dets = bboxes[remain_inds]
        scores_keep = scores[remain_inds]
        scores_second = scores[inds_second]

        if len(dets) > 0:
            '''Detections'''
            detections = [STrack(STrack.tlbr_to_tlwh(tlbr), s,frame_id=self.frame_id,id_in_frame=int(id_in_frame)) for
                          id_in_frame,(tlbr, s) in enumerate (zip(dets, scores_keep))]
            
        else:
            detections = []
        
        #####sophie mod: 
        for idx, det in enumerate(detections):
            det.id_in_frame=int(idx)
        
        ########

        ''' Add newly detected tracklets to tracked_stracks'''
        unconfirmed = []
        tracked_stracks = []  # type: list[STrack]
        for track in self.tracked_stracks:
            if not track.is_activated:
                unconfirmed.append(track)
            else:
                tracked_stracks.append(track)


        ''' Step 2: First association, with high score detection boxes'''
        strack_pool =  joint_stracks(tracked_stracks, self.lost_stracks) #tracked_stracks #sophie mod à gérer
        # Predict the current location with KF
        STrack.multi_predict(strack_pool)
        dists = matching.iou_distance(strack_pool, detections) #???? resultats bizarres à reverifier sophie: en ligne on a strack_pool et en colonnes on a detections
        ####sophie:mod
        
        self.DBN_current_track=copy.deepcopy(detections)        
        self.DBN_previous_track= copy.deepcopy(strack_pool)    
        self.DBN_dists = copy.deepcopy(dists)
        #### pour le moment on aura ps de U_second pck notre seuil est fixé à 0

        if not self.mot20:
            dists = matching.fuse_score(dists, detections) #sophie: on multiplie les iou par les scores de detection avant de faire le 1- fuse à nouveau
        self.DBN_dists = copy.deepcopy(dists)
       
        matches, u_track, u_detection = matching.linear_assignment(dists, thresh=self.match_thresh)
        ######sophie mod
        #print("***thres",self.match_thresh)
        for itracked, idet in matches:
            track = strack_pool[itracked]
            det = detections[idet]
            if track.state == TrackState.Tracked:
                track.update(detections[idet], self.frame_id,id_in_frame=idet)
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                refind_stracks.append(track)
        #c'est sophie qui l'ajoute 
        """for it in u_track:
            track = strack_pool[it]
            if not track.state == TrackState.Lost:
                track.mark_lost()
                lost_stracks.append(track)"""
                
        ''' Step 3: Second association, with low score detection boxes'''
        # association the untrack to the low score detections
        if len(dets_second) > 0:
            '''Detections'''
            detections_second = [STrack(STrack.tlbr_to_tlwh(tlbr), s,frame_id=self.frame_id,id_in_frame=int(id_in_frame)) for
                          id_in_frame,(tlbr, s) in enumerate (zip(dets_second, scores_second))] 
            #for
            #              (tlbr, s) in zip(dets_second, scores_second)]
        else:
            detections_second = []
        r_tracked_stracks = [strack_pool[i] for i in u_track if strack_pool[i].state == TrackState.Tracked]
        dists = matching.iou_distance(r_tracked_stracks, detections_second)
        matches, u_track, u_detection_second = matching.linear_assignment(dists, thresh=0.9) #sophie mod thresh=0.5
        for itracked, idet in matches:
            track = r_tracked_stracks[itracked]
            det = detections_second[idet]
            if track.state == TrackState.Tracked:
                track.update(det, self.frame_id, id_in_frame=idet)
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                refind_stracks.append(track)
        
        for it in u_track:
            track = r_tracked_stracks[it]
            if not track.state == TrackState.Lost:
                track.mark_lost()
                lost_stracks.append(track)
        
        '''Deal with unconfirmed tracks, usually tracks with only one beginning frame'''
        detections = [detections[i] for i in u_detection]
        dists = matching.iou_distance(unconfirmed, detections)
        if not self.mot20:
            dists = matching.fuse_score(dists, detections)
        matches, u_unconfirmed, u_detection = matching.linear_assignment(dists, thresh=0.9)#sophie mod thresh=0.7
        for itracked, idet in matches:
            unconfirmed[itracked].update(detections[idet], self.frame_id, id_in_frame=idet)
            activated_starcks.append(unconfirmed[itracked])
        for it in u_unconfirmed:
            track = unconfirmed[it]
            track.mark_removed()
            removed_stracks.append(track) 

        """ Step 4: Init new stracks"""
        if self.fixed_parc ==True:
            if self.frame_id<self.nbr_frames_fixing:
                print("initialisation of new tracks fixed for livestock context")

                for inew in u_detection:
                    track = detections[inew]
                    if track.score < self.det_thresh:
                        continue
                    track.activate(self.kalman_filter, self.frame_id)
                    activated_starcks.append(track)
                #print(track, "activated")
                #exit(0)
        else:
            print("initialisation of new tracks in variable context")
            for inew in u_detection:
                track = detections[inew]
                if track.score < self.det_thresh:
                    continue
                track.activate(self.kalman_filter, self.frame_id)
                activated_starcks.append(track)

                
        """ Step 5: Update state"""
        """if self.frame_id==4300:
            print("break")"""
        for track in self.lost_stracks:
            if self.frame_id - track.end_frame > self.max_time_lost:
                track.mark_removed()
                removed_stracks.append(track)

        # print('Ramained match {} s'.format(t4-t3))

        self.tracked_stracks = [t for t in self.tracked_stracks if t.state == TrackState.Tracked]
        self.tracked_stracks = joint_stracks(self.tracked_stracks, activated_starcks) # les nouveaux tracks activés 
        self.tracked_stracks = joint_stracks(self.tracked_stracks, refind_stracks) #sophie: les tracks retrouvés grâces aux petits scores de box. 
        self.lost_stracks = sub_stracks(self.lost_stracks, self.tracked_stracks)
        self.lost_stracks.extend(lost_stracks)
        self.lost_stracks = sub_stracks(self.lost_stracks, self.removed_stracks)
        self.removed_stracks.extend(removed_stracks)
        self.tracked_stracks, self.lost_stracks = remove_duplicate_stracks(self.tracked_stracks, self.lost_stracks)
        # get scores of lost tracks
        output_stracks = [track for track in self.tracked_stracks if track.is_activated]


        return output_stracks


def joint_stracks(tlista, tlistb):
    exists = {}
    res = []
    for t in tlista:
        exists[t.track_id] = 1
        res.append(t)
    for t in tlistb:
        tid = t.track_id
        if not exists.get(tid, 0):
            exists[tid] = 1
            res.append(t)
    return res


def sub_stracks(tlista, tlistb):
    stracks = {}
    for t in tlista:
        stracks[t.track_id] = t
    for t in tlistb:
        tid = t.track_id
        if stracks.get(tid, 0):
            del stracks[tid]
    return list(stracks.values())


def remove_duplicate_stracks(stracksa, stracksb):
    pdist = matching.iou_distance(stracksa, stracksb)
    pairs = np.where(pdist < 0.15)
    dupa, dupb = list(), list()
    for p, q in zip(*pairs):
        timep = stracksa[p].frame_id - stracksa[p].start_frame
        timeq = stracksb[q].frame_id - stracksb[q].start_frame
        if timep > timeq:
            dupb.append(q)
        else:
            dupa.append(p)
    resa = [t for i, t in enumerate(stracksa) if not i in dupa]
    resb = [t for i, t in enumerate(stracksb) if not i in dupb]
    return resa, resb
